"test.graphics" <-
function(funs=NULL,
         package=NULL,
         names=NULL,
         omit=NULL,
         width=600,
         height=600,
         device=postscript,
         format="pbm",
         model.loc=".",
         test.loc=model.loc,
         verbose=FALSE,
         quiet=FALSE,
         reset.rng=TRUE,
         # additional arguments to the device
         ...) {
  dev <- getDevice(device)
  # omit should be NULL, or char vector, or list
  if (!is.null(omit)) {
    if (is.character(omit))
      omit <- as.list(omit)
    if (!is.list(omit))
      stop("Invalid `omit' argument")
  }
  # format should be character
  format <- as.character(format)
  if (charmatch(format, "pbm", nomatch=FALSE))
    format.fns <- pbm.fns
  else if (charmatch(format, "ps", nomatch=FALSE))
    format.fns <- ps.fns
  else if (charmatch(format, "pdf", nomatch=FALSE))
    format.fns <- pdformat.fns
  else
    stop(paste("Output format", format, "has not yet been implemented"))
  # width and height must be integer
  width <- as.integer(width)
  height <- as.integer(height)
  # model.loc must be character
  model.loc <- as.character(model.loc)
  test.loc <- as.character(test.loc)
  funs <- getFuns(funs, package)
  # names must be NULL or same length as funs
  if (!is.null(names)) {
    if (length(names) != length(funs))
      stop("`names' must be the same length as `funs'")
  }
  nodiffs <- TRUE
  count <- 1
  diffs <- list()
  for (i in funs) {
    # Skip this function if it is named in a element of omit
    if (i %in% omit)
      next
    if (verbose)
      cat("test.graphics:", i, "\n");
    # Generate the test output
    testfilename <- paste(if (is.null(names)) i else names[count],
                          "-", dev$name, "-",
                          width, "x", height, sep="")
    fulltestfilename <- file.path(test.loc, testfilename)
    dev$fns$open(fulltestfilename, ".test.", width, height, ...)
    # Shut up example()'s warnings about "help file, but no examples"
    options(warn=-1)
    # Reset random number generator so that randomly-generated data
    # do not show up as differences
    if (reset.rng)
      set.seed(7654, "Wichmann-Hill")
    # Need to do something like rm(list = ls(all.names = TRUE))
    # to tidy up the environment between runs of example()
    # Wrap this in a try so that any "stop()"s in example code
    # do not chuck us out
    # Can't use example(i) because example() always does a substitute(topic)
    try(eval(call("example", i, echo=FALSE)))
    options(warn=0)
    dev$fns$close(fulltestfilename, ...)
    # Convert to bitmap format
    dev$fns$convert(test.loc, testfilename, ".test.",
                       format, width, height, ...)
    # Diff the test output against the model output
    # Can assume existence of diff on all platforms?
    modelfilename <- paste(if (is.null(names)) i else names[count],
                           "-", dev$name, "-",
                           width, "x", height, sep="")
    fullmodelfilename <- file.path(model.loc, modelfilename)
    # Get lists of files to compare
    # Is list.files sufficiently platform-independent?
    # Looks like it's designed to be
    # (see listfiles in $RHOME/src/main/platform.c)
    modelfiles <- list.files(model.loc,
                             paste("^", modelfilename, ".save.*.",
                                   format, sep=""),
                             full.names=TRUE)
    testfiles <- list.files(test.loc,
                            paste("^", testfilename, ".test.*.",
                                  format, sep=""),
                            full.names=TRUE)
    # Some sanity checks
    # NOTE this implicitly checks that bitmap size, bitmap format,
    # and number of examples is the same
    if (length(modelfiles) != length(testfiles)) {
      warning(paste("Number of model files (",
                    length(modelfiles),
                    ") does not match number of test files (",
                    length(testfiles), ") for function ", i, sep=""))
    }
    index <- 1
    for (j in modelfiles) {
      # Skip example if omit says so
      if (index %in% omit[[i]])
        next
      # No matter what format is produced, the comparison
      # occurs via "diff" (?)
      if (length(system(paste("diff", modelfiles[index], testfiles[index]),
                        TRUE))
          > 0) {
        nodiffs <- FALSE
        # Generate a list containing the differences
        if (is.null(diffs[[i]]))
          diffs[[i]] <- index
        else
          diffs[[i]] <- c(diffs[[i]], index)
        if (verbose)
          cat(paste("    DIFFERENCE in example", index, "\n"))
        format.fns$diff(modelfiles[index], testfiles[index], test.loc)
      }
      index <- index + 1
    }
    count <- count + 1
  }
  if (!quiet) {
    if (nodiffs)
      cat("************\nNo differences detected\n************\n")
    else
      cat("************\nDIFFERENCES DETECTED\n************\n")
  }
  invisible(diffs)
}
